@@ -17,27 +17,26 @@
 ** MA 02111-1307, USA
 */
 
-#include "davfsd.h"
-#include "webdav.h"
-#include "util.h"
+#include "config.h"
 
-#include <stdlib.h>
-#include <string.h>
+#include <error.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <signal.h>
+#include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
 
-/* Linux Kernel Source */
-/* avoid inclusion of kernel time.h, since it conflicts with userland time.h */
-#define _LINUX_TIME_H
-#include <linux/coda.h>
-
 /* Neon */
 #include <ne_basic.h>
 #include <ne_alloc.h>
+#include <ne_socket.h>
+
+#include "davfsd.h"
+#include "webdav.h"
+#include "util.h"
 
 /* Variables set at start up and needed for cleanup before exit */
 char *dav_url;
@@ -45,6 +44,9 @@ char *dav_mpoint;
 char *dav_pidfile;
 int dav_locks;
 
+/* File descriptor of coda device. */
+int dav_cfs;
+
 /* a linked list of the above objects */
 static davfs_list *davfs_list_root;
 
@@ -57,7 +59,7 @@ struct stat dav_dir_stat;
 int dav_noexec;*/
 
 int dav_has_permission(struct coda_in_hdr *ih) {
-#ifdef NEW_CODA_STRUCTURES
+#if CODA_KERNEL_VERSION > 2
     DBG1("id: %i\n", ih->uid);
     if ((ih->uid != dav_file_stat.st_uid) && (ih->uid != 0))
         return 0;
@@ -65,21 +67,40 @@ int dav_has_permission(struct coda_in_hdr *ih) {
     DBG1("id: %i\n", ih->cred.cr_uid);
     if ((ih->cred.cr_uid != dav_file_stat.st_uid) && (ih->cred.cr_uid != 0))
         return 0;
-#endif /* NEW_CODA_STRUCTURES */
+#endif /* CODA_KERNEL_VERSION */
     return 1;
 }
 
+static void put_all() {
+    DBG0("Saving open files on exit.");
+    davfs_list *node = davfs_list_root;
+    while (node != NULL) {
+        struct v_list *h = (struct v_list *) node->data;
+        if (h->has_changed && !h->is_dir && !h->is_created) {
+            if (dav_put_unlock(h->real_name, h->local_name) < 0)
+                syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                       "Error saving %s", h->real_name);
+        }
+        if (h->local_name != NULL)
+            remove(h->local_name);
+        node = node->next;
+    }
+}
+
+
 static void default_signal_handler(int signo)
 {
     DBG1("signal_handler(%d)\n", signo);
     signal(SIGTERM, SIG_IGN);
     signal(SIGHUP, SIG_IGN);
     signal(SIGINT, SIG_IGN);
+    put_all();
     remove(dav_pidfile);
     dav_close_connection();
     exit(EXIT_SUCCESS);
 }
 
+
 static void sigchld_signal_handler(int signo)
 {
     DBG1("signal_handler(%d)\n", signo);
@@ -87,6 +108,7 @@ static void sigchld_signal_handler(int signo)
     exit(EXIT_FAILURE);
 }
 
+
 static void dav_set_signal() {
     signal(SIGTERM, default_signal_handler);
     signal(SIGHUP, default_signal_handler);
@@ -101,72 +123,46 @@ static void dav_set_signal() {
 }
 
 /* Clear coda cache of a file entry */
-static void zap_file(int cfs, union inputArgs *in_buf)
+static void zap_file(union inputArgs *in_buf)
 {
     struct coda_zapfile_out cmd;
-    cmd.oh.opcode = CODA_ZAPFILE;
+    cmd.oh.opcode = CODA_PURGEFID;
     cmd.oh.unique = 0;
     cmd.CodaFid = in_buf->coda_close.VFid;
-    write(cfs, &cmd, sizeof(cmd));
+    write(dav_cfs, &cmd, sizeof(cmd));
 }
 
 
-/*
-** FIXME : need to be redesign
-** file name maintenance (chache) 
-**
-*/
 static char *look_name(DAVCodaFid *id, int *is_dir)
 {
     char *name;
 
     if ((id->Volume != MY_VOL) || (id->Vnode != MY_VNODE)) {
-#ifdef NEW_CODA_STRUCTURES
-        DBG3("Bad handle passed %lx/%lx/%lx\n", id->Volume, id->Vnode,
-             id->List);
-#else
-        DBG3("Bad handle passed %lx/%lx/%lx\n", id->Volume, id->Vnode,
-             id->Unique);
-#endif /* NEW_CODA_STRUCTURES */
+         syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                "Bad handle passed %x/%x/%p\n", id->Volume, id->Vnode,
+                 id->List);
         kill(getpid(), SIGTERM);
     }
     if (PARANOIA) {
         davfs_list *walk;
         for (walk = davfs_list_root; walk; walk = walk->next) {
-#ifdef NEW_CODA_STRUCTURES
             if (walk->data == id->List)
                 break;
-#else
-            if (walk->data == (void*) id->Unique)
-                break;
-#endif /* NEW_CODA_STRUCTURES */
         }
         if (!walk) {
-#ifdef NEW_CODA_STRUCTURES
-            printf("Bad handle passed %lx/%lx/%lx\n", id->Volume, id->Vnode,
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "Bad handle passed %x/%x/%p\n", id->Volume, id->Vnode,
                    id->List);
-#else
-            printf("Bad handle passed %lx/%lx/%lx\n", id->Volume, id->Vnode,
-                   id->Unique);
-#endif /* NEW_CODA_STRUCTURES */
             return NULL;    
         }
     }                /* paranoia */
-#ifdef NEW_CODA_STRUCTURES
     name = id->List->real_name;
     *is_dir = id->List->is_dir;
-#else
-    name =  ((struct v_list *) id->Unique)->real_name;
-    *is_dir =  ((struct v_list *) id->Unique)->is_dir;
-#endif /* NEW_CODA_STRUCTURES */
     
     return name;
 }
 
-/*
- * make a new file name and put it the list
- * FIXME: should we check whether the name is unique??
- */
+
 static void alloc_vfid(DAVCodaFid *res, char *name, int is_dir)
 {
     struct v_list *h = ne_calloc(sizeof(*h));
@@ -186,40 +182,19 @@ static void alloc_vfid(DAVCodaFid *res, char *name, int is_dir)
     if (res != NULL) {
         res->Volume = MY_VOL;
         res->Vnode = MY_VNODE;
-#ifdef NEW_CODA_STRUCTURES
         res->List = h;
-#ifdef DAVCODAFID_INIT_RESERVED
-        res->Reserved = 0;
-#endif
-#else
-        res->Unique = (u_int32_t)h;
-#endif /* NEW_CODA_STRUCTURES */
     }
     DBG3("New vfid created at %p (file <%s: %d>)\n", 
          h, h->real_name, h->is_dir);
 }
 
 
-/* 
-** Show the list foe DEBUG
-*/
-#ifdef DEBUG
-static void dump_vfid_list(void)
-{
-    davfs_list  *walk;
-    for (walk = davfs_list_root; walk; walk = walk->next) {
-        struct v_list *v = (struct v_list *) walk->data;
-        printf("list element <%s:%d>\n", v->real_name, v->is_dir);
-    }
-}
-#endif
-
 /*
 ** Delete name from the list 
 */
 static void delete_name(const char *real_name)
 {
-    davfs_list *walk, *prev, *to_free;
+    davfs_list *walk, *prev;
 
     DBG1("Delete_name %s\n", real_name);
     /* at this point, the paranoid would check for a dirty file and write it out */
@@ -292,38 +267,33 @@ static inline int CodaToVfsFlags(int flags)
 static inline char *get_local_copy(DAVCodaFid id, int flags)
 {
     int fd;
-#ifdef NEW_CODA_STRUCTURES
     struct v_list *v = id.List;
-#else
-    struct v_list *v = (struct v_list *) id.Unique;
-#endif /* NEW_CODA_STRUCTURES */
-    /* I would normally put an assert here, but look_name() was called already */
-    /* truncated files do not need an ftp read first; this code cluster handles it */
-    v->is_created = 0;
-    if (flags & C_O_WRITE)
-    v->has_changed = 1;
 
-    if (flags & C_O_TRUNC) {
-        v->is_trunc = 1;
+    if ((flags & C_O_TRUNC) || v->is_created){
         v->local_name = dav_get_tempnam("trunc", &fd);
-    
-        if(fd==-1) 
-            perror(v->local_name);
-        else {
-            /* FIXME : Do we need this? */
-            /*
-               setmode(fd, CodaToVfsFlags(flags));
-            */
+        if(fd==-1) {
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "Could not create temp file.");
+        } else {
+            DBG1("  cache-file: %s", v->local_name);
+            v->has_changed = 1;
+            v->is_created = 0;
+            v->is_trunc = 1;
             close(fd);
         }
         return v->local_name;
-    }
-    /* Read  or Write*/
-    else {
-        char *name = dav_getlocalcopy(v->real_name);
-        if (name == NULL)
-            fprintf(stderr, "\nget_local_copy: NULL for '%s'\n", v->real_name);
-        return v->local_name = name;
+
+    } else {
+        v->local_name = dav_getlocalcopy(v->real_name);
+        if (v->local_name == NULL) {
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "get_local_copy: NULL for '%s'\n", v->real_name);
+        } else {
+            DBG1("  cache-file: %s", v->local_name);
+            if (flags & C_O_WRITE)
+                v->has_changed = 1;
+        }
+        return v->local_name;
     }
 }
 
@@ -364,7 +334,7 @@ static void write_dir_entry(int handle, const char *name,
         /* FIXME: we should propagate error */
         perror("\a" PACKAGE_NAME ": error writing dirent");    
 
-    DBG3("['%s', handle=%x, ino=%lx]", name, handle, ent.d_fileno);
+    DBG3("['%s', handle=%x, ino=%x]", name, handle, ent.d_fileno);
 }
 
 /* Global variable for directory cache */
@@ -375,11 +345,7 @@ dav_prop_result *prev_list_result;
  * Save information for directory cache */
 static char *get_dir_copy(DAVCodaFid id)
 {
-#ifdef NEW_CODA_STRUCTURES
     struct v_list *v = id.List;
-#else
-    struct v_list *v = (struct v_list *) id.Unique;
-#endif /* NEW_CODA_STRUCTURES */
     dav_prop_result *one_res;
     dav_prop_result *result = dav_opendir(v->real_name);
     char *cache;
@@ -396,10 +362,6 @@ static char *get_dir_copy(DAVCodaFid id)
     prev_list_result = result;
     prev_base_dir = ne_strdup(v->real_name);
     dav_no_trail(prev_base_dir);
-    if(prev_base_dir==0) {
-        NE_FREE(prev_base_dir);
-        prev_base_dir==ne_strdup("/");
-    }
     /* It's directory */
     v->is_dir = 1;
 
@@ -460,44 +422,64 @@ static char *get_dir_copy(DAVCodaFid id)
     return v->local_name = cache;
 }
 
+static int sync_file(DAVCodaFid id) {
+    struct v_list *v = id.List;
+
+    if (v->is_dir)
+        return 0;
+
+    if (access(v->local_name, F_OK) != 0) {
+        DBG1("  %s: missing cache file on a fsync", v->real_name);
+        return EIO;
+    }
+
+    if (v->has_changed) {
+        DBG1("  saving the file %s ", v->local_name);
+        return dav_put(v->real_name, v->local_name);
+    }
+    
+    return 0;
+}
+
 static int unget_local_copy(DAVCodaFid id)
 {
-    int return_value = 0;
-#ifdef NEW_CODA_STRUCTURES
+    int ret = 0;
     struct v_list *v = id.List;
-#else
-    struct v_list *v = (struct v_list *) id.Unique;
-#endif /* NEW_CODA_STRUCTURES */
-    if (v->local_name) {
-        /* FIXME : Do we really need to check it? */
-        if (v->has_changed) {
-            DBG1("saving the file %s ", v->local_name);
-            return_value = dav_put_unlock(v->real_name, v->local_name);
-        }
-    
-        if (v->is_dir==1) {
-            dav_closedir(prev_list_result);
-            prev_list_result = NULL;
-            NE_FREE(prev_base_dir);
-        }
-#if 0
-        if (v->is_dir==0) {
-            unlink(v->local_name);
-        }
-    
-        /* Always put is trunc */
-        else if (v->is_trunc) {
 
-        }
-#endif
-        unlink(v->local_name);
-        NE_FREE(v->local_name);
+    DBG3("  unget_: u%i c%i cr%i", v->usage, v->has_changed, v->is_created);
+    if (v->local_name == NULL) {
+        DBG1("  %s: empty local_name on a close", v->real_name);
+        if (dav_locks && v->has_changed)
+            dav_unlock(v->real_name);
         v->has_changed = 0;
+        v->is_trunc = 0;
+        return EIO;
     }
-    else {
-        DBG0("<empty file name on a close>");
+
+    if (access(v->local_name, F_OK) != 0) {
+        DBG1("  %s: missing cache file on a close", v->real_name);
+        if (dav_locks && v->has_changed)
+            dav_unlock(v->real_name);
+        v->has_changed = 0;
+        v->is_trunc = 0;
+        return EIO;
     }
-    return return_value;
+
+    if (v->is_dir) {
+        dav_closedir(prev_list_result);
+        prev_list_result = NULL;
+        NE_FREE(prev_base_dir);
+    } else if (v->has_changed) {
+        DBG1("  saving the file %s ", v->local_name);
+        if (dav_put_unlock(v->real_name, v->local_name) < 0)
+            ret = dav_get_errno();
+    }
+    
+    DBG1("  removing %s ", v->local_name);
+    unlink(v->local_name);
+    NE_FREE(v->local_name);
+    v->has_changed = 0;
+    return ret;
 }
 
 /****************** attribute mangling *************************************/
@@ -548,11 +530,6 @@ static void st2attr(struct stat *s, struct coda_vattr *a)
     a->va_filerev = 0;
 }
 
-static inline int legal_name(char *name)
-{
-    return 1;
-}
-
 
 /* 
 ** General Open function 
@@ -561,49 +538,91 @@ static char *coda_open(char *name, DAVCodaFid id, int flags,
                union inputArgs *in_buf, union outputArgs *out_buf)
 {
     struct stat st;
+    struct v_list *v = id.List;
 
+    DBG3("  open: u%i c%i cr%i", v->usage, v->has_changed, v->is_created);
     if (!v_created(id)) {
-        if (!legal_name(name)) {
-            out_buf->oh.result = ENOENT;
-            return 0;
-        }
-    
-        /* It's file */
-        else if (dav_stat((name), &st, 0, 0) == -1) {
+        if (dav_stat((name), &st, 0, 0) == -1) {
             out_buf->oh.result = dav_get_errno();
-            DBG1("file '%s' probably does not exist.", (name));
+            if (out_buf->oh.result == ENOENT && strlen(name) > 0) {
+                DBG1("  file '%s' probably does not exist.", name);
+                delete_name(name);
+                zap_file(in_buf);
+            }
             return NULL;
         }
-
-    }
-    else
+    } else {
         st = dav_file_stat;
-
-    /* Directory */
-    if (S_ISDIR(st.st_mode)) {
-        DBG0("[open request on a directory]");
-        name = get_dir_copy(id);
     }
-    /* General file open */
-    else {
-        /* If is is write or trunc (NULL lock) */
-        if (flags & C_O_WRITE) {
-            if (dav_lock(name) && dav_locks == 1 ) { /* Lock fail?? */
-                DBG1("LOCK failed: %s\n", name);
+
+    if (S_ISDIR(st.st_mode)) {
+        DBG0("  open request on a directory");
+        if (v->local_name != NULL) {
+            if (access(v->local_name, F_OK) == 0)
+                remove(v->local_name);
+            NE_FREE(v->local_name);
+            v->usage = 0;
+            dav_closedir(prev_list_result);
+            prev_list_result = NULL;
+            NE_FREE(prev_base_dir);
+        }
+        get_dir_copy(id);
+        if (v->local_name == NULL) {
+            out_buf->oh.result = EIO;
+            return NULL;
+        } 
+
+    } else if (v->usage > 0) {
+        DBG0("  open request on already open file");
+        if (v->local_name == NULL) {
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "%s: no local_name but usage >0", name);
+            v->usage = 0;
+            v->has_changed = 0;
+            dav_unlock(name);
+            out_buf->oh.result = EIO;
+            return NULL;
+        }
+        if (access(v->local_name, F_OK) != 0) {
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "%s: missing cache file", name);
+            NE_FREE(v->local_name);
+            v->usage = 0;
+            v->has_changed = 0;
+            dav_unlock(name);
+            out_buf->oh.result = EIO;
+            return NULL;
+        }
+        if ((flags & C_O_WRITE) && dav_locks && !v->has_changed) {
+            if (dav_lock(name) < 0) {
+                syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                       "%s: LOCK failed", name);
                 out_buf->oh.result = dav_get_errno();
                 return NULL;
             }
         }
-    
-        name = get_local_copy(id, flags);
-        if (!name) {
+
+    } else {
+        if (flags & C_O_WRITE) {
+            if (dav_locks && dav_lock(name)) {
+                syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                       "%s: LOCK failed", name);
+                out_buf->oh.result = dav_get_errno();
+                return NULL;
+            }
+        }
+        get_local_copy(id, flags);
+        if (v->local_name == NULL) {
             out_buf->oh.result = dav_get_errno();
-            DBG1("local copy of  '%s' failed.", (name));
+            dav_unlock(name);
+            syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_ERR),
+                   "%s: failed to get local copy", name);
             return NULL;
         }
     }
-    v_usage(in_buf->coda_open.VFid) += 1;
-    return name;
+
+    v->usage++;
+    return v->local_name;
 }
 
 #define DAVFS_DEBUG \
@@ -633,11 +652,14 @@ int main(int argc, char *argv[]) {
     dav_check_config(args);
 
     DBG1("Open device %s", args->device);
-    int cfs = open(args->device, O_RDWR, 0);    
-    if (cfs < 0)
-        error(EXIT_FAILURE, 0, "Could not open %s", args->device);
+    dav_cfs = open(args->device, O_RDWR, 0);    
+    if (dav_cfs < 0)
+        error(EXIT_FAILURE, 0,
+              "Could not open %s:\n"
+              "\tKernel does not support coda,"
+              "or the coda-module is not loaded.", args->device);
     u_int coda_kernel_version;
-    ioctl(cfs, CIOC_KERNEL_VERSION, &coda_kernel_version);
+    ioctl(dav_cfs, CIOC_KERNEL_VERSION, &coda_kernel_version);
     if (coda_kernel_version != CODA_KERNEL_VERSION)
         error(EXIT_FAILURE, 0, "CODA kernel version mismatch:\n"
                           "\tcompiled for version %d, version %d available",
@@ -648,7 +670,7 @@ int main(int argc, char *argv[]) {
     union inputArgs *in_buf = ne_calloc(SIZEOF_INBUF);
     union outputArgs *out_buf = ne_calloc(SIZEOF_OUTBUF);
     if ((in_buf == NULL) || (out_buf == NULL))
-        error(EXIT_FAILURE, 0, "can't allocate message buffers");
+        error(EXIT_FAILURE, 0, "Can't allocate message buffers.");
         
     DBG1("Open connection to %s", dav_url);    
 #ifdef DEBUG
@@ -657,19 +679,19 @@ int main(int argc, char *argv[]) {
     if (ne_sock_init())
         error(EXIT_FAILURE, 0, "Socket library initalization failed");
     if (dav_open_connection(args) < 0)
-        error(EXIT_FAILURE, 0, "Could not connect to %s", dav_url);
+        error(EXIT_FAILURE, 0, "Could not connect to %s.", dav_url);
 
     DBG0("Create root VFid");
     struct stat st;
-    alloc_vfid(NULL, "/", 1);
-    dav_stat("/", &st, 1, 0);
+    alloc_vfid(NULL, "", 1);
+    dav_stat("", &st, 1, 0);
 
     DBG0("Fork into daemon mode; parent will mount");
     dav_set_signal();
     pid_t childpid = fork();
     if (childpid > 0) {
         /* Parent */
-        if (dav_mount(cfs, args) != 0) {
+        if (dav_mount(dav_cfs, args) != 0) {
             /* Fails. Kill child pid and exit */
             kill(childpid, SIGTERM);
             exit(EXIT_FAILURE);
@@ -677,7 +699,7 @@ int main(int argc, char *argv[]) {
         DBG1("Parent leaves now. Good luck to child no. %i!", childpid);
         exit(EXIT_SUCCESS);
     } else if (childpid < 0) {
-        error(EXIT_FAILURE, 0, "Could not start child process for daemon");
+        error(EXIT_FAILURE, 0, "Could not start child process for daemon.");
     }
 
     DBG0("Child: No way back to root privileges.\n");
@@ -693,6 +715,8 @@ int main(int argc, char *argv[]) {
     setsid();
 
     for ( ; ; ) {
+    /* Look out: This variable is really mad. It sometimes points to
+       real_name, sometimes points to local_name, hoping to get you confused. */
     char *name;
     int msg, size, retval;
     fd_set rfds;
@@ -701,11 +725,11 @@ int main(int argc, char *argv[]) {
     int is_dir;
 
     FD_ZERO(&rfds);
-    FD_SET(cfs, &rfds);
+    FD_SET(dav_cfs, &rfds);
     tv.tv_sec = SELECT_TIME_OUT;
     tv.tv_usec = 0;
 
-    retval = select(cfs + 1, &rfds, NULL, NULL, &tv);
+    retval = select(dav_cfs + 1, &rfds, NULL, NULL, &tv);
     if (retval == 0) {
         DBG1("Waited for %i s\n", SELECT_TIME_OUT);
         if (!dav_still_mounted())
@@ -717,11 +741,11 @@ int main(int argc, char *argv[]) {
 
     memset(in_buf, 0, SIZEOF_INBUF);
     memset(out_buf, 0, SIZEOF_OUTBUF);
-    msg = read(cfs, in_buf, (SIZEOF_INBUF - 1));
+    msg = read(dav_cfs, in_buf, (SIZEOF_INBUF - 1));
     if ((msg == -1) || (msg == 0))
         continue;
 
-    DBG2("got %3.3d byte command: opcode = %2.2ld ", msg,
+    DBG2("got %3.3d byte command: opcode = %2.2d ", msg,
             in_buf->ih.opcode);
     
     out_buf->oh.opcode = in_buf->ih.opcode;
@@ -740,17 +764,10 @@ int main(int argc, char *argv[]) {
         while (node->next != NULL) {
             node = node->next;
         }
+        memset(&out_buf->coda_root.VFid, 0, sizeof(DAVCodaFid));
         ((DAVCodaFid *)&out_buf->coda_root.VFid)->Volume = MY_VOL;
         ((DAVCodaFid *)&out_buf->coda_root.VFid)->Vnode = MY_VNODE;
-#ifdef NEW_CODA_STRUCTURES
         ((DAVCodaFid *)&out_buf->coda_root.VFid)->List = (struct v_list*) node->data;
-#ifdef DAVCODAFID_INIT_RESERVED
-        ((DAVCodaFid *)&out_buf->coda_root.VFid)->Reserved = 0;
-#endif
-#else
-        ((DAVCodaFid *)&out_buf->coda_root.VFid)->Unique = (u_int32_t) node->data;
-#endif /* NEW_CODA_STRUCTURES */
-        v_usage(out_buf->coda_root.VFid) = 1;
         break;
         
         /********************* file props & access *******************/
@@ -778,14 +795,16 @@ int main(int argc, char *argv[]) {
         CMD_NOREP(setattr);
         DUMP(setattr);
         GET_VFID_name(setattr);
-        DBG3
-          (" try to set mode to %o, mdate to %lx and owner to %d: ",
-           in_buf->coda_setattr.attr.va_mode,
-           in_buf->coda_setattr.attr.va_mtime.tv_sec,
-           in_buf->coda_setattr.attr.va_uid);
+        DBG3(" m:%o, o:%d, g:%d",
+             in_buf->coda_setattr.attr.va_mode,
+             in_buf->coda_setattr.attr.va_uid,
+             in_buf->coda_setattr.attr.va_gid);
+        DBG3(" at:%lu, mt:%lu, ct:%lu",
+             in_buf->coda_setattr.attr.va_atime.tv_sec,
+             in_buf->coda_setattr.attr.va_mtime.tv_sec,
+             in_buf->coda_setattr.attr.va_ctime.tv_sec);
         /*2005-05-29, werner, security fix*/
-        if ((in_buf->coda_setattr.attr.va_mode != (u_short) -1)
-                || (in_buf->coda_setattr.attr.va_uid != -1)
+        if ((in_buf->coda_setattr.attr.va_uid != -1)
                 || (in_buf->coda_setattr.attr.va_gid != -1))
             /*dav_chmod( name, in_buf->coda_setattr.attr.va_mode );*/
             out_buf->oh.result = ENOTSUP;
@@ -910,8 +929,26 @@ int main(int argc, char *argv[]) {
         }
         break;
 
-    /*# FIXME : RELEASE */
     case CODA_RELEASE:
+        if (!dav_has_permission(&in_buf->ih)) {
+            out_buf->oh.result = EPERM;
+            break;
+        }
+        CMD_NOREP(release);
+        DUMP(release);
+        GET_VFID_name(release);
+        v_usage(in_buf->coda_close.VFid)--;
+        if (v_usage(in_buf->coda_close.VFid) < 0)
+            v_usage(in_buf->coda_close.VFid) = 0;
+        if (v_usage(in_buf->coda_close.VFid) == 0) {
+            out_buf->oh.result =
+                    unget_local_copy(*(DAVCodaFid *)&in_buf->coda_close.VFid);
+            if (out_buf->oh.result != 0)
+                zap_file(in_buf);
+        }
+        DBG1(" the current VFid list has %d elements\n", count);
+        break;
+        
     case CODA_CLOSE:
         if (!dav_has_permission(&in_buf->ih)) {
             out_buf->oh.result = EPERM;
@@ -920,13 +957,16 @@ int main(int argc, char *argv[]) {
         CMD_NOREP(close);
         DUMP(close);
         GET_VFID_name(close);
-        DBG0("CLOSE");
-        if (unget_local_copy(*(DAVCodaFid *)&in_buf->coda_close.VFid) < 0) {
-            out_buf->oh.result = dav_get_errno();
-            zap_file(cfs, in_buf);
+        v_usage(in_buf->coda_close.VFid)--;
+        if (v_usage(in_buf->coda_close.VFid) < 0)
+            v_usage(in_buf->coda_close.VFid) = 0;
+        if (v_usage(in_buf->coda_close.VFid) == 0) {
+            out_buf->oh.result =
+                    unget_local_copy(*(DAVCodaFid *)&in_buf->coda_close.VFid);
+            if (out_buf->oh.result != 0)
+                zap_file(in_buf);
         }
         DBG1(" the current VFid list has %d elements\n", count);
-        v_usage(in_buf->coda_close.VFid) -= 1;
         break;
         
         
@@ -937,6 +977,8 @@ int main(int argc, char *argv[]) {
             break;
         }
         CMD_NOREP(fsync);
+        DUMP(fsync);
+        out_buf->oh.result = sync_file(*(DAVCodaFid *)&in_buf->coda_close.VFid);
         break;
 
     case CODA_CREATE:
@@ -1080,10 +1122,10 @@ int main(int argc, char *argv[]) {
         break;
     }
     
-    DBG3(" returning %s (%ld), %d bytes\n\n",
+    DBG3(" returning %s (%d), %d bytes\n\n",
             out_buf->oh.result ? "FAILURE" : "SUCCESS",
             out_buf->oh.result, size);
-    msg = write(cfs, out_buf, size);
+    msg = write(dav_cfs, out_buf, size);
     if (toclose != -1)
         close(toclose);
     fflush(stdout);
