@@ -17,29 +17,39 @@
 ** MA 02111-1307, USA
 */
 
-#include "util.h"
+#include "config.h"
 
-#include <stdlib.h>
-#include <stdio.h>
+#include <argz.h>
+#include <ctype.h>
+#include <error.h>
 #include <errno.h>
 #include <fstab.h>
 #include <getopt.h>
 #include <grp.h>
 #include <mntent.h>
 #include <pwd.h>
-#include <signal.h> /* needed for linux/coda.h, don't know why */
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <termios.h>
+#include <unistd.h>
+#include <sys/mount.h>
+#include <sys/stat.h>
 
 /* Linux coda head file */
 /* avoid inclusion of kernel time.h, since it conflicts with userland time.h */
 #define _LINUX_TIME_H
+/* This macro is used in coda.h and causes compile errors.
+   We don't need it and define it as empty. */
+#define __user
 #include <linux/coda.h>
 
 /* Neon */
 #include <ne_alloc.h>
 #include <ne_string.h>
 
+#include "util.h"
+
 
 /* Global variables defined in davfsd.c */
 /* ==================================== */
@@ -50,7 +60,7 @@ extern int dav_locks;
 extern struct stat dav_file_stat;
 extern struct stat dav_dir_stat;
 /*2005-05-29, werner, security fix
-extern int dav_noexec;/*
+extern int dav_noexec;*/
 
 
 /* Global variables defined in webdav.c */
@@ -78,16 +88,17 @@ char *dav_mounts[3] = { DAV_PROC_MOUNTS, _PATH_MOUNTED, NULL };
     DBG1("    gid = %i", args->gid); \
     DBG1("    file_mode = %o", args->file_mode); \
     DBG1("    dir_mode = %o", args->dir_mode); */ \
-    DBG1("    ro = %i", args->ro); \
-    DBG1("    nodev = %i", args->nodev); \
-    DBG1("    noexec = %i", args->noexec); \
-    DBG1("    nosuid = %i", args->nosuid); \
+    DBG1("    ro = %lu", args->ro); \
+    DBG1("    nodev = %lu", args->nodev); \
+    DBG1("    noexec = %lu", args->noexec); \
+    DBG1("    nosuid = %lu", args->nosuid); \
     DBG1("    user = %i", args->user); \
     DBG1("    p_host = %s", args->p_host); \
     DBG1("    p_port = %i", args->p_port); \
     DBG1("    askauth = %i", args->askauth); \
     DBG1("    useproxy = %i", args->useproxy); \
     DBG1("    locks = %i", args->locks); \
+    DBG1("    mountanyway = %i", args->mountanyway); \
 } while (0)
 
 #define DBG_VARS(x) do { \
@@ -188,9 +199,10 @@ void usage(const char *prog) {
     printf("       users        : same as user\n");
     printf("       proxy=       : name and -optional- port of the http-proxy\n");
     printf("       [no]useproxy : (don't) use proxy for connection\n");
-    printf("       [no]askauth  : (don't) ask interactivly for creditentials if not found\n");
+    printf("       [no]askauth  : (don't) ask interactivly for credentials if not found\n");
     printf("                      in the secrets files.\n");
     printf("       [no]locks    : (don't) lock resources on the server\n\n");
+    printf("       [no]mountanyway: (don't) mount even without connection\n\n");
 }
 
 
@@ -205,17 +217,9 @@ void usage(const char *prog) {
 int arg_to_int(const char *arg, int base, const char *opt) {
     char *tail = NULL;
     int n = strtol(arg, &tail, base);
-    if ((n < 0) || (tail == NULL)) {
-        char *b = NULL;
-        if (base == 10) {
-            b = " decimal";
-        } else if (base == 8) {
-            b == "n octal";
-        } else {
-            b = "";
-        }
-        error(EXIT_FAILURE, 0, "Argument must be a%s number -%s", b, opt);
-    }
+    if ((n < 0) || (tail == NULL))
+        error(EXIT_FAILURE, 0, "Argument must be a base-%i-number -%s", base,
+              opt);
     return n;
 }
 
@@ -264,6 +268,8 @@ void get_options(dav_args *args, char *option) {
         NOASKAUTH,
         LOCKS,
         NOLOCKS,
+        MOUNTANYWAY,
+        NOMOUNTANYWAY,
         RO,
         RW,
         NODEV,
@@ -279,7 +285,7 @@ void get_options(dav_args *args, char *option) {
         DEFAULTS,
         END
     };
-    const char *suboptions[] = {
+    char* suboptions[] = {
         /*2005-05-29, werner, security fix, fixed mode
         [UID] = "uid",
         [GID] = "gid",
@@ -292,6 +298,8 @@ void get_options(dav_args *args, char *option) {
         [NOASKAUTH] = "noaskauth",
         [LOCKS] = "locks",
         [NOLOCKS] = "nolocks",
+        [MOUNTANYWAY] = "mountanyway",
+        [NOMOUNTANYWAY] = "nomountanyway",
         [RO] = "ro",
         [RW] = "rw",
         [NODEV] = "nodev",
@@ -310,8 +318,6 @@ void get_options(dav_args *args, char *option) {
     
     int so;
     char *argument = NULL;
-    struct passwd *pwd;
-    struct group *grp;
 
     while (*option != 0) {
         so = getsubopt(&option, suboptions, &argument);
@@ -321,7 +327,7 @@ void get_options(dav_args *args, char *option) {
         switch (so) {
         /*2005-05-29, werner, security fix, fixed mode
         case UID:
-            pwd = getpwnam(argument);
+            struct passwd *pwd = getpwnam(argument);
             if (pwd == NULL) {
                 args->uid = arg_to_int(argument, 10, suboptions[so]);
             } else {
@@ -329,7 +335,7 @@ void get_options(dav_args *args, char *option) {
             }
             break;
         case GID:
-            grp = getgrnam(argument);
+            struct group *grp = getgrnam(argument);
             if (grp == NULL) {
                 args->gid = arg_to_int(argument, 10, suboptions[so]);
             } else {
@@ -363,6 +369,12 @@ void get_options(dav_args *args, char *option) {
         case NOLOCKS:
             args->locks = 0;
             break;
+        case MOUNTANYWAY:
+            args->mountanyway = 1;
+            break;
+        case NOMOUNTANYWAY:
+            args->mountanyway = 0;
+            break;
         case RO:
             args->ro = MS_RDONLY;
             break;
@@ -430,6 +442,8 @@ void check_fstab(const dav_args *args) {
     dav_args *n_args = new_args();
     setfsent();
     struct fstab *ft = getfsfile(dav_mpoint);
+    if (ft == NULL)
+        ft = getfsfile(basename(dav_mpoint));
     if (ft != NULL) {
         err = 0;
         err |= strcmp(dav_url, ft->fs_spec);
@@ -448,6 +462,7 @@ void check_fstab(const dav_args *args) {
         err |= (args->useproxy != n_args->useproxy);
         err |= (args->askauth != n_args->askauth);
         err |= (args->locks != n_args->locks);
+        err |= (args->mountanyway != n_args->mountanyway);
     }
     endfsent();
     dav_delete_args(n_args);
@@ -494,7 +509,7 @@ char *dev_from_options(char *option) {
         DEVICE,
         END
     };
-    const char *suboptions[] = {
+    char *suboptions[] = {
         [DEV] = "dev",
         [DEVICE] = "device",
         [END] = NULL
@@ -567,10 +582,10 @@ int get_free_device(char **device) {
         if(avail[minor] == 0) {
             if (asprintf(device, "%s%i", DAV_DEV_NAME, minor) < 0)
                 abort();
-            dav_pidfile = ne_concat(STATEDIR, "/",
+            dav_pidfile = ne_concat(DAV_STATEDIR, "/",
                                     basename(*device), ".pid", NULL);
             if (access(dav_pidfile, F_OK) == 0) {
-                NE_FREE(device);
+                NE_FREE(*device);
                 NE_FREE(dav_pidfile);
             } else {
                 return minor;
@@ -578,6 +593,7 @@ int get_free_device(char **device) {
         }
     }
     error(EXIT_FAILURE, 0, "No free davfs-device to mount");
+    return 0;
 }
 
 
@@ -629,7 +645,6 @@ int parse_line(char *line, int parmc, char *parmv[]) {
         END
     };
     int state = SPACE;
-    int error = 0;
     int parm_no = 0;
     char buf[254];
     char *pos = buf;
@@ -757,7 +772,7 @@ int read_secrets(dav_args *args, int user) {
         return -1;
     }
     
-    int n = 0;
+    size_t n = 0;
     char *line = NULL;
     int length = getline(&line, &n, file);
     int lineno = 1;
@@ -824,7 +839,7 @@ int read_config(dav_args *args, int user) {
         return -1;
     }
     
-    int n = 0;
+    size_t n = 0;
     char *line = NULL;
     int length = getline(&line, &n, file);
     int lineno = 1;
@@ -896,14 +911,14 @@ char *ask_password() {
 }
 
 
-/* Get creditentials for proxy or WebDAV-server from user interactivly-
+/* Get credentials for proxy or WebDAV-server from user interactivly-
  * user     : pointer to a string pointer, where a pointer to the newly
  *            allocated user string will be stored
  * passowrd : pointer to a string pointer, where a pointer to the newly
  *            allocated passowrd string will be stored
  * kind     : should be "server" or "proxy" to inform the user
  * url      : the url of the server or the FQDN of the proxy the
- *            creditentials are needed for (to inform the user) */
+ *            credentials are needed for (to inform the user) */
 void ask_auth(char **user, char **password, const char *kind, const char *url) {
     if (*user == NULL) {
         printf("Please enter the username for authentication with %s\n"
@@ -1010,7 +1025,7 @@ dav_args *dav_get_args(int argc, char *argv[]) {
             dav_url = ne_strdup(argv[i]);
         }
         i++;
-        dav_mpoint = (char*) canonicalize_file_name(argv[i]);
+        dav_mpoint = canonicalize_file_name(argv[i]);
         break;
     default:
         printf("Too many arguments.\n");
@@ -1039,16 +1054,20 @@ void dav_check_config(dav_args *args) {
 
     if (getuid() != 0) {
         DBG0("  Checking group membership");
-        struct passwd *pw = getpwuid(getuid());
         struct group *grp = getgrnam(DAV_GROUP);
         if (grp == NULL)
             error(EXIT_FAILURE, 0, "Group %s does not exist.", DAV_GROUP);
-        char **members = grp->gr_mem;
-        int is_member = 0;
-        while (*members && !is_member) {
-            if (strcmp(*members, pw->pw_name) == 0)
-                is_member = 1;
-            members++;
+        int is_member = (getgid() == grp->gr_gid);
+        if (!is_member) {
+            int ngroups = getgroups(0, NULL);
+            gid_t *groups = (gid_t *) ne_calloc(ngroups * sizeof(gid_t));
+            ngroups = getgroups(ngroups, groups);
+            int i = 0;
+            while (!is_member && (i < ngroups)) {
+                is_member = (*(groups + i) == grp->gr_gid);
+                ++i;
+            }
+            NE_FREE(groups);
         }
         if (!is_member)
             error(EXIT_FAILURE, 0, "You must be member of group %s",
@@ -1088,7 +1107,7 @@ void dav_check_config(dav_args *args) {
             args->useproxy = 0;
     }
 
-    DBG0("  Looking up creditentials");
+    DBG0("  Looking up credentials");
     if (getuid() != 0)
         read_secrets(args, 1);
     seteuid(0);
@@ -1179,6 +1198,22 @@ int dav_mount(int fd, const dav_args *args) {
 
 
 void dav_save_mount_pid(void) {
+    if (access(DAV_STATEDIR, F_OK) != 0) {
+        seteuid(0);
+        DBG1("  Creating %s", DAV_STATEDIR);
+        if (mkdir(DAV_STATEDIR,
+                  S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH | S_ISVTX) != 0)
+            error(EXIT_FAILURE, 0, "Could not create %s", DAV_STATEDIR);
+        struct group *grp = getgrnam(DAV_GROUP);
+        if (grp == NULL)
+            error(EXIT_FAILURE, 0, "Group %s does not exist.", DAV_GROUP);
+        if (chown(DAV_STATEDIR, 0, grp->gr_gid) != 0)
+            error(EXIT_FAILURE, 0, "Could not change group of %s", DAV_STATEDIR);
+        if (chmod(DAV_STATEDIR,
+                  S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH | S_ISVTX) != 0)
+            error(EXIT_FAILURE, 0, "Could not change mode of %s", DAV_STATEDIR);
+        seteuid(getuid());
+    }
     DBG1("  Writing pid-file %s", dav_pidfile);
     FILE *fp = fopen(dav_pidfile, "w");
     if (fp == NULL)
@@ -1217,7 +1252,6 @@ void dav_no_trail(char *str) {
 
 
 char *dav_get_tempnam(const char *prefix, int *fd) {
-    int i;
     char tmp[1024];
     snprintf(tmp, sizeof(tmp)-1, "%s/%s_%s_XXXXXX",
              DAV_TMP_DIR, PACKAGE_NAME, prefix);
