@@ -17,22 +17,27 @@
 ** MA 02111-1307, USA
 */
 
-#include "webdav.h"
+#include "config.h"
 
 #include <errno.h>
 #include <fcntl.h>
 #include <stdio.h>
-#include <unistd.h>
+#include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include <ne_auth.h>
 #include <ne_basic.h>
+#include <ne_dates.h>
 #include <ne_locks.h>
 #include <ne_props.h>
 #include <ne_request.h>
 #include <ne_socket.h>
 #include <ne_uri.h>
 
+#include "webdav.h"
+
+
 extern struct stat dav_file_stat;
 extern struct stat dav_dir_stat;
 /*2005-05-29, werner, security fix
@@ -51,12 +56,14 @@ static ne_uri server = { 0 };        /* current URI of server. */
 
 static char *path;            /* current working collection */
 
-static char *pnt;
-
 static ne_lock_store *lock_store;
 
 static int dav_errno; /* dav error number */
 
+/* When > 0 this indicates to mount even without connection, must
+   be reset when the filesystem is mounted. */
+int mount_anyway = 0;
+
 
 #define STATUS(x)  (atoi(ne_get_error(session))==(x))
 
@@ -95,6 +102,7 @@ void dav_close_connection(void)
 static int dav_set_errno() {
     int http_status=500;
     const char *err_msg = ne_get_error(session);
+    DBG1("NE_ERROR: %s", err_msg);
     
     /* parse the status line */
     if (err_msg)
@@ -138,10 +146,10 @@ int dav_get_errno() {
     return dav_errno;
 }
 
-/* Call-back-function called by neon to get creditentials.*/
+/* Call-back-function called by neon to get credentials.*/
 static int auth(void *ud, const char *realm, int attempt,
                 char *username, char *password) {
-    DBG2("Neon wants creditentials for %s. %i. attempt.\n", realm, attempt);
+    DBG2("Neon wants credentials for %s. %i. attempt.\n", realm, attempt);
     if (dav_username != NULL)
         strncpy(username, dav_username, NE_ABUFSIZ - 1);
     if (dav_password != NULL)
@@ -149,10 +157,10 @@ static int auth(void *ud, const char *realm, int attempt,
     return attempt;
 }
 
-/* Call-back-function called by neon to get creditentials.*/
+/* Call-back-function called by neon to get credentials.*/
 static int proxy_auth(void *ud, const char *realm, int attempt,
                       char *username, char *password) {
-    DBG2("Neon wants creditentials for %s. %i. attempt.\n", realm, attempt);
+    DBG2("Neon wants credentials for %s. %i. attempt.\n", realm, attempt);
     if (dav_p_user != NULL)
         strncpy(username, dav_p_user, NE_ABUFSIZ - 1);
     if (dav_p_password != NULL)
@@ -213,9 +221,15 @@ static int dav_init_session(dav_args *args) {
 
     /* Set timeout */
     ne_set_read_timeout(session, DAV_READ_TIMEOUT);
+
     
+#if NE_VERSION_MINOR == 24
     ne_set_expect100(session, 1);
-    ne_set_useragent(session, PACKAGE_STRING);
+#endif
+    
+    char *useragent = ne_concat(PACKAGE_TARNAME, "/", PACKAGE_VERSION, NULL);
+    ne_set_useragent(session, useragent);
+    NE_FREE(useragent);
     
     ne_hook_pre_send(session, set_close_head, NULL);
     
@@ -243,39 +257,28 @@ int dav_open_connection(dav_args *args) {
 
     server.path = NULL;        /* this is used as a temporary store. */
     
-    /* Single argument: see whether we have a path or scheme */
-    /*if (strchr(dav_url, '/') == NULL) {*/
-    if (false) {
-        /* No path, no scheme -> just a hostname */
-        pnt = strchr(dav_url, ':');
-        if (pnt != NULL) {
-            *pnt++ = '\0';
-            server.port = atoi(pnt);
-        }
-        else {
-            server.port = 80;
-        }
-        server.host = ne_strdup(dav_url);
-        server.scheme = ne_strdup("http");
-    }
-    else {
     /* Parse the URL */
-        if (ne_uri_parse(dav_url, &server) || server.host == NULL) {
-            fprintf(stderr, "Could not parse URL `%s'\n", dav_url);
-            return -1;
-        }
-    
-        if (server.scheme == NULL)
-            server.scheme = ne_strdup("http");
-    
-        if (!server.port)
-            server.port = ne_uri_defaultport(server.scheme);
+    if (ne_uri_parse(dav_url, &server) || server.host == NULL) {
+        fprintf(stderr, "Could not parse URL `%s'\n", dav_url);
+        return -1;
+    }
+    if( strlen(server.path)>1 && *(server.path+strlen(server.path)-1)=='/')
+            *(server.path+strlen(server.path)-1)=0;
 
-        if (strcasecmp(server.scheme, "https") == 0) {
-            if (!ne_supports_ssl()) {
-                fprintf(stderr, "SSL is not enabled.\n");
-                return -1;
-            }
+    if (server.scheme == NULL)
+        server.scheme = ne_strdup("http");
+
+    if (!server.port)
+        server.port = ne_uri_defaultport(server.scheme);
+
+    if (strcasecmp(server.scheme, "https") == 0) {
+#if NE_VERSION_MINOR == 24
+        if (!ne_supports_ssl()) {
+#else
+        if (!ne_has_support(NE_FEATURE_SSL)) {
+#endif
+            fprintf(stderr, "SSL is not enabled.\n");
+            return -1;
         }
     }
 
@@ -285,29 +288,62 @@ int dav_open_connection(dav_args *args) {
     dav_init_session(args);
 
     /* Ugly workaround for http-escaped urls and non-http-escaped
-       urls with spaces. Blame the people that need spaces and
-       other special characters in pathnames. */
+       urls with spaces. */
     char *tmp_path = ne_path_unescape(server.path);
     NE_FREE(server.path);
     server.path = tmp_path;
-    tmp_path = ne_path_escape(server.path);
+
+    if (strcmp(server.path, "/") == 0) {
+        tmp_path = ne_strdup("/");
+    } else {
+        char *p = ne_concat(server.path, "/", NULL);
+        tmp_path = ne_path_escape(p);
+        NE_FREE(p);
+    }
     ret = ne_options(session, tmp_path, &caps);
     NE_FREE(tmp_path);
+    path = server.path;
 
     switch (ret) {
     case NE_OK:
-        path = server.path;
         break;
     
     case NE_CONNECT:
         if (args->p_host) {
-            fprintf(stderr, "Could not connect to `%s' on port %d:\n%s\n",
+            fprintf(stderr, "Could not connect to `%s' on port %d:\n  %s\n",
                     args->p_host, args->p_port, ne_get_error(session));
         } 
         else {
-            fprintf(stderr, "Could not connect to `%s' on port %d:\n%s\n",
+            fprintf(stderr, "Could not connect to `%s' on port %d:\n  %s\n",
                     server.host, server.port, ne_get_error(session));
         }
+
+        if (args->mountanyway) {
+                mount_anyway = 2;
+                DBG0("Mounting whithout connection to the server.");
+                fprintf(stderr, "Mounting whithout connection to the server.\n");
+                return 0;
+        } else {
+            char *s = NULL;
+            size_t n = 0;
+            ssize_t len = 0;
+            printf("Do you want to mount the filesystem anyway?\n");
+            printf("It will only be useable when the server becomes"
+                   " reachable again.\n");
+            printf("Mount anyway [y,N]? "),
+            len = getline(&s, &n, stdin);
+            if (len < 0)
+                abort();
+            if (*(s + len -1) == '\n')
+                *(s + len -1) = '\0';
+            if (strcmp(s, "y") == 0) {
+                NE_FREE(s);
+                mount_anyway = 2;
+                DBG0("Mounting whithout connection to the server.");
+                return 0;
+            }
+            NE_FREE(s);
+        }
         break;
     
     case NE_LOOKUP:
@@ -320,7 +356,6 @@ int dav_open_connection(dav_args *args) {
         break;
     }
    
-DBG1("   path: %s", path);
     return dav_set_errno();
 }
 
@@ -378,7 +413,6 @@ static char *resolve_path(const char *p, const char *filename, int isdir)
     }
     
     escaped_uri = ne_path_escape(ret);
-    DBG1("WebDAV URL: %s\n", ret);
     
     NE_FREE(path);
     NE_FREE(fname);
@@ -390,7 +424,7 @@ static char *resolve_path(const char *p, const char *filename, int isdir)
 /*
 ** callback hander to save results
 */
-static void results(void *userdata, const char *uri,
+static void results(void *userdata, const char *o_uri,
             const ne_prop_result_set * set)
 {
     dav_prop_result_header *hdr = userdata;
@@ -399,14 +433,29 @@ static void results(void *userdata, const char *uri,
     const char *data;
     int is_dir = 0;
 
+    /* o_uri may be a compete url or only a path.
+       We need the path. */
+    const char *uri;
+    char *url = ne_uri_unparse(&server);
+    *(url + strlen(url) - strlen(server.path)) = '\0';
+    if (strstr(o_uri, url) == o_uri) {
+        uri = o_uri + strlen(url);
+    } else {
+        uri = o_uri;
+    }
+    NE_FREE(url);
+
     /* Allocation for one result */
     result = ne_calloc(sizeof(*result));
 
     /* displayname */
     data = ne_propset_value(set, &stat_props[0]);
-    if (data) 
+    if (data) {
         result->fname = ne_path_unescape(data);
-    else {
+      	if( strlen(result->fname)>0
+      	                && result->fname[strlen(result->fname)-1]=='/')
+    		    result->fname[strlen(result->fname)-1]=0;
+    } else {
         char *parent = ne_path_parent(uri);
         char *fname;
         /* It is / */
@@ -418,7 +467,7 @@ static void results(void *userdata, const char *uri,
             while(*fname=='/')
                 fname++;
         
-            if( strlen(fname)>1 && *(fname+strlen(fname)-1)=='/')
+            if( strlen(fname)>0 && *(fname+strlen(fname)-1)=='/')
                 *(fname+strlen(fname)-1)=0;
 
             result->fname = ne_strdup(fname); //TODO: free me
@@ -519,7 +568,8 @@ dav_prop_result * dav_opendir(const char *name)
     ne_propfind_handler *pfh;
     
     phdr.uri  = resolve_path(path, name, 1);
-    
+
+    DBG1("  PROPFIND 1 %s", phdr.uri);
     pfh = ne_propfind_create(session, phdr.uri, 1);
     ret = ne_propfind_named(pfh, stat_props, results, &phdr);
     
@@ -541,6 +591,7 @@ void dav_closedir(dav_prop_result *res)
 int dav_mkdir(char *name, int mode)
 {
     char *uri = resolve_path(path, name, 0);
+    DBG1("  MAKECOLL %s", uri);
     ne_mkcol(session, uri);
 
     NE_FREE(uri);
@@ -551,6 +602,7 @@ int dav_rmdir(char *name)
 {
     char *uri = resolve_path(path, name, 1);
     
+    DBG1("  DELETE %s", uri);
     ne_delete(session, uri);
     NE_FREE(uri);
     
@@ -562,6 +614,7 @@ int dav_delete(char *name)
 {
     char *uri = resolve_path(path, name, 0);
     
+    DBG1("  DELETE %s", uri);
     ne_delete(session, uri);
     NE_FREE(uri);
     
@@ -577,6 +630,7 @@ int dav_rename(char *src, char *des, int is_dir, int try_num)
     char *des_uri = resolve_path(path, des, is_dir);
     
     /* Move with over write */
+    DBG2("  MOVE %s %s", src_uri, des_uri);
     ne_move(session, 1, src_uri, des_uri); 
     NE_FREE(src_uri);
     NE_FREE(des_uri);
@@ -601,13 +655,16 @@ char *dav_getlocalcopy(const char *name) {
         return NULL;
     }
 
-    /* FIXME : Do we need this? */
-    /*
-    setmode(fd, O_CREAT | O_TRUNC | O_WRONLY);
-    */
-
-    ne_get(session, uri, fd);
-    close(fd);
+    DBG1("  GET %s", uri);
+    if (ne_get(session, uri, fd) != NE_OK) {
+        dav_set_errno();
+        close(fd);
+        remove(fname);
+        NE_FREE(fname);
+    } else {
+        close(fd);
+    }
+        
     
     NE_FREE(uri);
 
@@ -618,7 +675,17 @@ char *dav_getlocalcopy(const char *name) {
 /* Lock the resource for write  and save the lock to lock_store */
 int dav_lock(const char *name) {
     char *uri = resolve_path(path, name, 0);
-    struct ne_lock *lock;
+    struct ne_lock *lock = NULL;
+    
+    /* Check whether already locked */
+    char *backup_path = server.path;
+    server.path = uri;
+    lock = ne_lockstore_findbyuri(lock_store, &server);
+    server.path = backup_path;
+    if (lock != NULL) {
+        NE_FREE(uri);
+        return 0;
+    }
     
     /* Let's lock the resource */
     lock = ne_lock_create();
@@ -629,9 +696,10 @@ int dav_lock(const char *name) {
     lock->type =  ne_locktype_write;
     lock->owner = ne_concat("<href>", "davfs2", "</href>", NULL);
     
+    DBG1("  LOCK %s", uri);
     NE_FREE(uri);
-    
-    if (ne_lock(session, lock)!=NE_OK) {
+
+    if (ne_lock(session, lock) != NE_OK) {
         /* otherwise, throw it away */
         ne_lock_destroy(lock);
         return dav_set_errno();
@@ -643,6 +711,57 @@ int dav_lock(const char *name) {
 }
 
 
+int dav_unlock(const char *name) {
+    int ret = NE_OK;
+    char *uri = resolve_path(path, name, 0);
+    struct ne_lock *lock;
+    
+    /* Lst's get the lock */
+    char *backup_path = server.path;
+    server.path = uri; /* Change path */
+    lock = ne_lockstore_findbyuri(lock_store, &server);
+    server.path = backup_path;
+    NE_FREE(uri);
+
+    if (!lock) {
+        DBG0("  Should be locked, before open\n");
+        return 0;
+    }
+    
+    DBG1("  LOCK %s", lock->uri.path);
+    if (ne_unlock(session, lock) != NE_OK) {
+        ret = -1;
+    }
+    
+    ne_lockstore_remove(lock_store, lock);
+    ne_lock_destroy(lock);
+    return ret;
+}
+
+
+int dav_put(const char *name, const char *fname) {
+    int fd;
+    int ret = NE_OK;
+    char *uri = resolve_path(path, name, 0);
+
+    fd = open(fname, O_RDONLY);
+    if( fd < 0 ) {
+        NE_FREE(uri);
+        return EIO;
+    }
+
+    DBG1("  PUT %s", uri);
+    if( ne_put(session, uri, fd) != NE_OK) { 
+        dav_set_errno();
+        ret = dav_get_errno();
+    }
+    close(fd);
+
+    NE_FREE(uri);
+    return ret;
+}
+
+
 /* Get localfile and save it to the server 
  * It should be locked
  */
@@ -654,11 +773,13 @@ int dav_put_unlock(const char *name, const char *fname) {
     
     fd = open(fname, O_RDONLY);
     
-    /* FIXME : Should I return error */
-    if( fd ==-1 ) {
+    fd = open(fname, O_RDONLY);
+    if( fd < 0 ) {
+        NE_FREE(uri);
+        return -1;
     }
-    
 
+    DBG1("  PUT %s", uri);
     if( ne_put(session, uri, fd)!=NE_OK) { 
         dav_set_errno();
         ret = -1;
@@ -666,16 +787,19 @@ int dav_put_unlock(const char *name, const char *fname) {
     close(fd);
     
     /* Lst's get the lock */
-    server.path = (char *)uri; /* Change path */
+    char *backup_path = server.path;
+    server.path = uri; /* Change path */
     lock = ne_lockstore_findbyuri(lock_store, &server);
+    server.path = backup_path;
     NE_FREE(uri);
 
     if (!lock) {
-        DBG0("Shuld be locked, before open\n");
-        return dav_set_errno();
+        DBG0("  Should be locked, before open\n");
+        return -1;
     }
     
-    if (ne_unlock(session, lock)!=NE_OK) {
+    DBG1("  UNLOCK %s", lock->uri.path);
+    if (ne_unlock(session, lock) != NE_OK) {
         dav_set_errno();
         ret = -1;
     }
@@ -711,7 +835,7 @@ int dav_stat_cache(const char *name, struct stat *st,
         dav_no_trail(parent);
     }
    
-    DBG2("\nCACHE DIR: %s<->%s\n", parent, 
+    DBG2("\n  CACHE DIR: %s<->%s\n", parent, 
          (base_dir?base_dir:"(null)"));
            
     /* Diffrent parent. Let's return */
@@ -725,11 +849,11 @@ int dav_stat_cache(const char *name, struct stat *st,
     
     /* Let's walk and find the file is in the cache */
     for (one_res = result; one_res; one_res= one_res->next) {
-        DBG2("\nCACHE FNAME: %s<->%s\n", 
+        DBG2("\n  CACHE FNAME: %s<->%s\n", 
              one_res->fname, name+len);
         /* Is the same file name */
         if (!strcmp(one_res->fname, name+len)) {
-            DBG0("I got it from cache \n");
+            DBG0("  I got it from cache \n");
             *st = one_res->f_st;
             return NE_OK;
         }
@@ -753,6 +877,7 @@ int dav_stat(const char *name, struct stat *st,
     /* It is stat */
     phdr.is_stat = 1;
 
+    DBG1("  PROPFIND 0 %s", phdr.uri);
     pfh = ne_propfind_create(session, phdr.uri, 0);
     ret = ne_propfind_named(pfh, stat_props, results, &phdr);
 
@@ -769,6 +894,13 @@ int dav_stat(const char *name, struct stat *st,
     NE_FREE(phdr.uri);
     ne_propfind_destroy(pfh);
 
+    if (mount_anyway > 0) {
+        --mount_anyway;
+        if (ret != 0)
+            *st = dav_dir_stat;
+        return 0;
+    }
+
     switch(status) {
     case 405:
         fprintf(stderr, "WebDAV Enabled??\n");
