@@ -20,7 +20,6 @@
 #ifndef __WEBDAV_H__
 #define __WEBDAV_H__
 
-#include "config.h"
 #include "util.h"
 
 /* boolean */
@@ -92,6 +91,10 @@ int dav_rename(char *buf, char *buf2, int is_dir, int try_num);
 /* Lock the file */
 int dav_lock(const char *name);
 
+int dav_unlock(const char *name);
+
+int dav_put(const char *name, const char *fname);
+
 /* Get localfile and save it to the server */
 int dav_put_unlock(const char *name, const char *fname);
 
